--[[
    Diddy Kong Racing Self-Magnet Item by Altiami (formerly yoshimo)
    For xItemLib by minenice

	v0.4:
		(unlisted in v0.3) updated my screen name info in file top description
		MAGNET OVERHAUL
			Magnet no longer immediately pulls target back
			Magnetism is now thrown toward target; must hit target to trigger effect
			Item now does something even if there's no target
			Projectile has several properties and interactions
				Slightly homes in on target, moreso if target has Thunder Shield
				Electricity effect indicates if something is trying to home in on you
				Projectile (tries to) bounce off walls like Orbinaut Spikeballs
				Projectile interacts with springs, spring sectors, and death sectors
				Projectile expires after some time
				Projectile slows down to stop after a while before expiring
				Anyone can hit projectile. User can hit projectile to reclaim the item
				Projectile is damaged by usual sources
				Projectile detaches a shield item upon hitting it as if racer was hit
				Projectile hits through Thunder Shield and removes it
					Using Thunder Shield can still destroy the projectile
		Racers no longer sometimes stop suddenly after magnet dragging finishes
		Targetting no longer targets players waiting for comeback mode in battle
		Updated tips to match new behavior
		Fixed magnet mobjs to not be encore mapped
		Extra paranoia checks for targeting validity
		Adjusted electricity to "move" instead of having position set
		Fixed interpolation on electricity
		Updated namespace names for clarity between item and mod namespaces
		Fixed name for xItemLib from "X-Item Lib" in file top description

	v0.3:
		Added tap mode for players used to SRB2K's method of item activation
		Changes reticle thinker logic to hopefully fix ghost reticles
    
    v0.2.1:
        Fixed forgetting to change a function for new implementation
    
    v0.2:
        Fixed places where checking for player being alive didn't check if player is not respawning
        Same as above but the inverse
        Fixed attempting to iterate over magnet info without checking if nil
        Fixed magnet info not updating while conditions to stop targeting were met
    
    v0.1.1:
        Fixed player variable initialization for players that join after add-on load during a match
    
    v0.1:
        Initial
]]

freeslot(
    "MT_DKRRETICLEFRAME",
    "MT_DKRRETICLEGLASS",
    "MT_DKRMAGNETELECTRICITY",
	"MT_DKRMAGNETPROJECTILE",
    "sfx_atdrmg"
)

--cache enumerated global constants
--commented vars are only ever referenced once and are for future reference
--local ANG15 = ANG15
--local ANGLE_315 = ANGLE_315
local FRACBITS = FRACBITS
local MT_PLAYER = MT_PLAYER
local TICRATE = TICRATE
local k_invincibilitytimer = k_invincibilitytimer
local k_growshrinktimer = k_growshrinktimer
local PST_LIVE = PST_LIVE
local k_hyudorotimer= k_hyudorotimer
local k_respawn = k_respawn
local MT_DKRRETICLEFRAME = MT_DKRRETICLEFRAME
local MT_DKRRETICLEGLASS = MT_DKRRETICLEGLASS
local MT_DKRMAGNETELECTRICITY = MT_DKRMAGNETELECTRICITY
local k_itemamount = k_itemamount
local ANGLE_90 = ANGLE_90
local MFE_VERTICALFLIP = MFE_VERTICALFLIP
local sfx_atdrmg = sfx_atdrmg
--local SPR_ADSM = SPR_ADSM
local MT_FIREDITEM = MT_FIREDITEM
local FRACUNIT = FRACUNIT
local MT_DKRMAGNETPROJECTILE = MT_DKRMAGNETPROJECTILE
local MT_ITEMCLASH = MT_ITEMCLASH
local MF2_DONTDRAW = MF2_DONTDRAW
local ANGLE_270 = ANGLE_270
local k_curshield = k_curshield
local ANGLE_180 = ANGLE_180
local k_comebacktimer = k_comebacktimer
local k_bumper = k_bumper
local k_stealingtimer = k_stealingtimer
local k_stolentimer = k_stolentimer
local k_rocketsneakertimer = k_rocketsneakertimer
local k_eggmanexplode = k_eggmanexplode
local k_itemtype = k_itemtype
local sfx_itpick = sfx_itpick
local MT_ORBINAUT = MT_ORBINAUT
local MT_JAWZ = MT_JAWZ
local MT_JAWZ_DUD = MT_JAWZ_DUD
local MT_BANANA = MT_BANANA
local MT_BALLHOG = MT_BALLHOG
-- I AM SO TIRED OF THIS
local MT_SSMINE = MT_SSMINE
local MT_SSMINE_SHIELD = MT_SSMINE_SHIELD
do
	local tmp = MT_SSMINE
	MT_SSMINE = MT_SSMINE_SHIELD
	MT_SSMINE_SHIELD = tmp
end
local MT_ORBINAUT_SHIELD = MT_ORBINAUT_SHIELD
local MT_JAWZ_SHIELD = MT_JAWZ_SHIELD
local MT_BANANA_SHIELD = MT_BANANA_SHIELD
local MF_NOCLIPTHING = MF_NOCLIPTHING
local MF2_AMBUSH = MF2_AMBUSH
local MFE_UNDERWATER = MFE_UNDERWATER
local k_bananadrag = k_bananadrag
local k_itemheld = k_itemheld
local KITEM_NONE = KITEM_NONE
--local KITEM_THUNDERSHIELD = KITEM_THUNDERSHIELD
--local RING_DIST = RING_DIST

-- constants in source not exposed to Lua
local BOXTOP = 0
local BOXBOTTOM = 1
local BOXLEFT = 2
local BOXRIGHT = 3

-- not in source but useful as a constant
local THUNDER_SHIELD_DAMAGE_RADIUS = RING_DIST / 4

local HOLD_ACTIVATION_TIME_TICS = TICRATE / 5
--must be [180, 360) to account for 180 degrees always being negative in Lua
local TARGETING_FORCE_ANGLE_THRES = -ANG15
local TARGETING_ANY_ANGLE_THRES = ANGLE_315

local TARGETING_DIST_THRES = 1024 << FRACBITS
local MAGNET_KILL_DIST_THRES = mobjinfo[MT_PLAYER].radius << 3
local MAGNET_TARGET_OFFSET_DIST = mobjinfo[MT_PLAYER].radius << 2

local MAGNET_PULL_FORCE = 40 << FRACBITS
local MAGNET_TIME = TICRATE << 1
local MAGNET_ELECTRICITY_COUNT = 8
local MAGNET_ELECTRICITY_OFFSET = 8 << FRACBITS
local MAGNET_SOUND_LOOP = TICRATE * 712 / 2205 --approximation, truncated
local MAGNET_PROJECTILE_LAUNCH_DIFFERENCE_MINIMUM = 8 << FRACBITS
local MAGNET_PROJECTILE_LIFETIME = TICRATE << 2
local MAGNET_PROJECTILE_FORWARD_TIME = TICRATE
local MAGNET_PROJECTILE_DECELERATION = 2 << FRACBITS
local MAGNET_PROJECTILE_TURN_SPEED_MAX = ANGLE_45 / 15 -- 3 degrees
local MAGNET_PROJECTILE_TURN_SPEED_MAX_ON_THUNDER_SHIELD = MAGNET_PROJECTILE_TURN_SPEED_MAX << 2
local MAGNET_PROJECTILE_RADIUS = 16 << FRACBITS

local DKR_BACK_MAGNET_ITEM_NAMESPACE = "ITEM_DKR_BACK_MAGNET"
local DKR_BACK_MAGNET_MOD_NAMESPACE = "MOD_DKR_BACK_MAGNET"
local DKR_BACK_MAGNET_ID

-- I love hardcoded values; see k_kart.c:3159
local MAGNET_PROJECTILE_SPEED_MULT_FOR_KART_SPEED_LOOKUP = {
	[0] = FRACUNIT - (FRACUNIT >> 2),
	FRACUNIT,
	FRACUNIT + (FRACUNIT >> 2)
}

-- per-race scaled vars
local targetingDistThresScaled
local magnetKillDistThresScaled
local magnetTargetOffsetDistScaled
local magnetPullForceScaled
local magnetProjectileSpeedMult

local defrostingDebounce = false
-- used to bypass mobjs.iterate() for SPEED
local dkrMagnetProjectiles = {}

local function netSyncLocals(network)
    targetingDistThresScaled = network(targetingDistThresScaled)
    magnetKillDistThresScaled = network(magnetKillDistThresScaled)
    magnetPullForceScaled = network(magnetPullForceScaled)
    magnetTargetOffsetDistScaled = network(magnetTargetOffsetDistScaled)
	magnetProjectileSpeedMult = network(magnetProjectileSpeedMult)
    defrostingDebounce = network(defrostingDebounce)
	dkrMagnetProjectiles = network(dkrMagnetProjectiles)
end
addHook("NetVars", netSyncLocals)

local function getMapObjectScaled(a)
    return FixedMul(a, mapobjectscale)
end

local function ternaryOperation(c, t, f)
    local r = f
    if c then
        r = t
    end
    return r
end

local function randomInt()
    return (P_RandomFixed() << FRACBITS) | P_RandomFixed()
end

local function makeLoopingSoundData(sound, loopTime)
	return {
		sound = sound,
		loopTime = loopTime,
		timer = 0
	}
end

local function updateLoopingSound(soundData, origin)
	if soundData.timer == soundData.loopTime then
        --kill a sound and replay it
        S_StopSoundByID(origin, soundData.sound)
        S_StartSound(origin, soundData.sound)
        --1, not 0, to account for being == to loop is technically 0
        soundData.timer = 1
    else
        if not soundData.timer then
            --only play a new sound
            S_StartSound(origin, soundData.sound)
        end
        soundData.timer = $ + 1
    end
end

--thanks call
--inserted here by mine
local function generateHFForPlayer(p)
    if p and p.valid and p.mo and p.mo.valid then
        return {{p.name, p.skincolor}, {skins[p.mo.skin].facemmap, p.skincolor}, p}
    else
        return {{}, {}}
    end
end

local function playerKillsMagnetProjectile(playerMobj)
	return playerMobj.player and
		playerMobj.player.valid and
		(
			playerMobj.player.kartstuff[k_invincibilitytimer] > 0 or
		    playerMobj.player.kartstuff[k_growshrinktimer] > 0
		)
end

local function canDKRBackMagnetPlayer(player)
    return player
        and player.valid
        and player.mo
        and player.mo.valid
        and not player.spectator
        and player.playerstate == PST_LIVE
		and not (G_BattleGametype() and player.kartstuff[k_comebacktimer])
		and not player.kartstuff[k_respawn]
		and not playerKillsMagnetProjectile(player.mo)
end

local function canDKRTargetPlayer(player, otherP)
    return player
		and player.valid
		and player.mo
		and player.mo.valid
		and player ~= otherP
        and canDKRBackMagnetPlayer(otherP)
        and not otherP.kartstuff[k_hyudorotimer]
end

local function isMagnetAtInvalidRange(player, target)
    return R_PointToDist2(player.mo.x, player.mo.y, target.mo.x, target.mo.y)
            <= player.mo.radius + magnetKillDistThresScaled + target.mo.radius
        --[[or -abs(R_PointToAngle2(player.mo.x, player.mo.y, target.mo.x, target.mo.y))
            < ANGLE_270 + ANG20]]
end

local function mustKillAllMagnets(player)
    return
        not (player.mo and player.mo.valid) 
        or player.spectator
        or player.playerstate ~= PST_LIVE
        or player.kartstuff[k_respawn] ~= 0
end

local function isMagnetTargetValid(target)
    return
        not (target and target.valid and target.mo and target.mo.valid)
        or target.spectator
        or target.playerstate ~= PST_LIVE
        or target.kartstuff[k_respawn] ~= 0
end

local function relocateTargetPiece(piece, x, y, z, oldTarget)
    --reposition reticle
    P_MoveOrigin(piece, x, y, z)
    
    --INTERP: trick it into moving with old target if swapping targets
    if target == oldTarget then
        --makes it default to interp to new position
        piece.momx = 0
        piece.momy = 0
        piece.momz = 0
    elseif
        oldTarget
        and oldTarget.valid
        and oldTarget.mo
        and oldTarget.mo.valid
        and not oldTarget.spectator
    then
        local oldTargetMo = oldTarget.mo
        local x = oldTargetMo.momx
        local y = oldTargetMo.momy
        local z = oldTargetMo.momz
        if x or y or z then
            piece.momx = x
            piece.momy = y
            piece.momz = z
        else
            --trick it into barely moving at all for this tic
            piece.momz = 1
        end
    else
        --trick it into barely moving at all for this tic
        piece.momz = 1
    end
end

local function refreshDKRTargetStuff(player)
    local playerMo = player.mo
    local target
    local closestAngle = TARGETING_FORCE_ANGLE_THRES - 1
    local closestDist = targetingDistThresScaled + 1
    
    for i = 0, #players - 1 do
        local otherP = players[i]
        
        if canDKRTargetPlayer(player, otherP) then
            local dist = R_PointToDist2(playerMo.x, playerMo.y, otherP.mo.x, otherP.mo.y)
            
            if dist <= targetingDistThresScaled then
                local angle = -abs(R_PointToAngle2(
                    playerMo.x,
                    playerMo.y,
                    otherP.mo.x,
                    otherP.mo.y
                ) - playerMo.angle)
                
                if angle > closestAngle then
                    target = otherP
                    closestAngle = angle
                    --since someone is within the "scope" range, we only prioritize angle now
                    for j = i + 1, #players - 1 do
                        otherP = players[i]
                        if canDKRTargetPlayer(player, otherP) then
                            dist = R_PointToDist2(playerMo.x, playerMo.y, otherP.mo.x, otherP.mo.y)
                            if dist <= targetingDistThresScaled then
                                angle = -abs(R_PointToAngle2(
                                    playerMo.x,
                                    playerMo.y,
                                    otherP.mo.x,
                                    otherP.mo.y
                                ) - playerMo.angle)
                                if angle > closestAngle then
                                    target = otherP
                                    closestAngle = angle
                                end
                            end
                        end
                    end
                    break
                --never target players behind
                elseif angle > TARGETING_ANY_ANGLE_THRES and dist < closestDist then
                    target = otherP
                    closestDist = dist
                end
            end
        end
    end
    
    --print(target and ("target " .. target.name) or "no target")
    if target then
        local retiX = target.mo.x
        local retiY = target.mo.y
        local retiZ = target.mo.z + (target.mo.height >> 1)
        if player.dkrtargetreticle.frame and player.dkrtargetreticle.frame.valid then
            relocateTargetPiece(
                player.dkrtargetreticle.frame,
                retiX,
                retiY,
                retiZ,
                player.dkrtarget
            )
        else
            --spawn and position the reticle
            player.dkrtargetreticle.frame = P_SpawnMobj(retiX, retiY, retiZ, MT_DKRRETICLEFRAME)
            --allow it to trace back in case something weird happens
            player.dkrtargetreticle.frame.tracer = playerMo
        end
        if player.dkrtargetreticle.glass and player.dkrtargetreticle.glass.valid then
            relocateTargetPiece(
                player.dkrtargetreticle.glass,
                retiX,
                retiY,
                retiZ,
                player.dkrtarget
            )
        else
            --spawn and position the reticle
            player.dkrtargetreticle.glass = P_SpawnMobj(retiX, retiY, retiZ, MT_DKRRETICLEGLASS)
            --allow it to trace back in case something weird happens
            player.dkrtargetreticle.glass.tracer = playerMo
        end
        --update reticle color
        player.dkrtargetreticle.frame.color = player.skincolor
    else
        if player.dkrtargetreticle.frame and player.dkrtargetreticle.frame.valid then
            P_RemoveMobj(player.dkrtargetreticle.frame)
        end
        if player.dkrtargetreticle.glass and player.dkrtargetreticle.glass.valid then
            P_RemoveMobj(player.dkrtargetreticle.glass)
        end
        player.dkrtargetreticle = {}
    end
    player.dkrtarget = target
end

local function preuseDKRBackMagnet(p, cmd, ticsHeld, attackJustDown)
	if ticsHeld > HOLD_ACTIVATION_TIME_TICS then
		p.dkrmagnetprimed = true
	end
end

local function activateDKRBackMagnet(p)
	-- loosely based on K_SpawnKartMissile
	local magnetProjectile = P_SpawnMobj(
		p.mo.x + p.mo.momx,
		p.mo.y + p.mo.momy,
		p.mo.z,
		MT_DKRMAGNETPROJECTILE
	)

	local desiredSpeed = FixedMul(mobjinfo[MT_DKRMAGNETPROJECTILE].speed, magnetProjectileSpeedMult)
	-- if player is going very fast, use player's momentum
	if desiredSpeed - p.speed < MAGNET_PROJECTILE_LAUNCH_DIFFERENCE_MINIMUM then
		desiredSpeed = p.speed + MAGNET_PROJECTILE_LAUNCH_DIFFERENCE_MINIMUM
	end

    local target = p.dkrtarget
	local projectileAngle
    if canDKRTargetPlayer(p, target) then
		-- send toward target
		projectileAngle = R_PointToAngle2(p.mo.x, p.mo.y, target.mo.x, target.mo.y)
		-- prjectile needs to know what to slightly home into
		-- based on Jawz code
		magnetProjectile.tracer = target.mo
	else
		-- send out anyway as a feedback mechanic, like Jawz
		projectileAngle = p.mo.angle
    end

	local xAngleScale = cos(projectileAngle)
	local yAngleScale = sin(projectileAngle)
	local projectileSpeedX = FixedMul(desiredSpeed, xAngleScale)
	local projectileSpeedY = FixedMul(desiredSpeed, yAngleScale)
	magnetProjectile.target = p.mo -- based on source for parity purposes
	magnetProjectile.targetPlayer = p -- also store player to prevent dead mobj oversights
	magnetProjectile.threshold = 10 -- prevent picking it back up immediately
	-- spawn forward a little
	-- pretty please don't collide with anything for a bit big mcthankies
	magnetProjectile.flags = $ | MF_NOCLIPTHING
	P_SetOrigin(
		magnetProjectile,
		magnetProjectile.x + projectileSpeedX,
		magnetProjectile.y + projectileSpeedY,
		magnetProjectile.z
	)
	-- prevent flying off of downward slopes (thanks Ashnal)
	if P_IsObjectOnGround(p.mo) then
		-- forcefully ground the projectile
		if p.mo.eflags & MFE_VERTICALFLIP then
			magnetProjectile.z = magnetProjectile.ceilingz - magnetProjectile.height
		else
			magnetProjectile.z = magnetProjectile.floorz
		end
		-- set standingslope with a fake move
		P_TryMove(magnetProjectile, magnetProjectile.x, magnetProjectile.y, true)
	end
	magnetProjectile.flags = $ & ~MF_NOCLIPTHING
	magnetProjectile.angle = projectileAngle
	magnetProjectile.momx, magnetProjectile.momy = projectileSpeedX, projectileSpeedY
	-- store speed for future referece for adjusting momentum
	magnetProjectile.movefactor = desiredSpeed
	-- play firing sound
	S_StartSound(p.mo, magnetProjectile.info.seesound)
	-- "Go get 'em!"
	K_PlayAttackTaunt(p.mo)

	-- fired item effect
	-- spawn offsets directly based on source in P_MobjThinker
	local firingEffectOffset = p.mo.radius + mobjinfo[MT_FIREDITEM].radius
	local firingEffect = P_SpawnMobj(
		p.mo.x + FixedMul(firingEffectOffset, xAngleScale),
		p.mo.y + FixedMul(firingEffectOffset, yAngleScale),
		p.mo.z + p.mo.height / 3,
		MT_FIREDITEM
	)
	-- set up to follow according to shootables like orbi/jawz correctly
	firingEffect.movecount = 1
	firingEffect.movedir = p.mo.angle - projectileAngle
	firingEffect.target = p.mo

	-- reset targetting stuff
    p.dkrtarget = nil
    if p.dkrtargetreticle.frame and p.dkrtargetreticle.frame.valid then
        P_RemoveMobj(p.dkrtargetreticle.frame)
    end
    if p.dkrtargetreticle.glass and p.dkrtargetreticle.glass.valid then
        P_RemoveMobj(p.dkrtargetreticle.glass)
    end
    p.dkrtargetreticle = {}
	-- actually use the item lol
    p.kartstuff[k_itemamount] = $ - 1
	p.dkrmagnetprimed = false
end

local function useDKRBackMagnet(p, cmd)
	if p.dkrmagnetprimed then
		activateDKRBackMagnet(p)
	else
		p.dkrmagnetprimed = true
	end
end

local function killMagnet(player, i)
    for j = 1, MAGNET_ELECTRICITY_COUNT do
        if player.dkrmagnetisminfo[i].electricity[j].valid then
            P_RemoveMobj(player.dkrmagnetisminfo[i].electricity[j])
        end
    end
    table.remove(player.dkrmagnetisminfo, i)
end

local function resetPlayerVars(player)
    player.dkrtarget = nil
    player.dkrtargetreticle = {}
    player.dkrmagnetisminfo = {}
	player.dkrmagnetprimed = false
end

local function updateMagnetismElectricityObject(elec)
	elec.angle = randomInt()
    if P_RandomKey(2) then
        elec.eflags = $ | MFE_VERTICALFLIP
    else
        elec.eflags = $ & ~MFE_VERTICALFLIP
    end
end

local function updateMagnetDrags(mobj)
    local player = mobj.player
    
    local didDrag = false
    local magnetMomX = 0
    local magnetMomY = 0
    
    if mustKillAllMagnets(player) then
        for i = #player.dkrmagnetisminfo, 1, -1 do
            killMagnet(player, i)
        end
        return
    end
    --EXIT EARLY
    
    --iterate backwards to be able to remove entries during iteration
    for i = #player.dkrmagnetisminfo, 1, -1 do
        local magnetInfo = player.dkrmagnetisminfo[i]
        local magnetTarget = magnetInfo.target
        
        if
            isMagnetTargetValid(magnetTarget)
            --timer check moved up here to prevent electricity being removed too early
            or magnetInfo.timer == magnetInfo.timeUp
            --never kill the magnet on the first tic b/c of distance so it at least does something
            or magnetInfo.timer
                and isMagnetAtInvalidRange(player, magnetTarget)
        then
            killMagnet(player, i)
        elseif canDKRBackMagnetPlayer(player) then
            local magnetMo = magnetTarget.mo
            --magnetangrespect should always be either the dragged or the target
            local magnetRespectMo = magnetInfo.angRespect.mo
            local destX, destY = magnetMo.x, magnetMo.y
            --aim for one side depending on which side is closer
            local relAng = R_PointToAngle2(mobj.x, mobj.y, destX, destY) - magnetRespectMo.angle
            --positive -> [0, 180) deg -> left of player -> right of target
            --negative -> [180, 360) deg -> right of player -> left of target
            local angOff
            
            if relAng < 0 then
                angOff = magnetRespectMo.angle + ANGLE_90
            else
                angOff = magnetRespectMo.angle - ANGLE_90
            end
            
            local sideOff = mobj.radius + magnetTargetOffsetDistScaled + magnetMo.radius
            
            destX = $ + FixedMul(sideOff, cos(angOff))
            destY = $ + FixedMul(sideOff, sin(angOff))
            
            local finalAng = R_PointToAngle2(mobj.x, mobj.y, destX, destY)
            
            didDrag = true
            magnetMomX = $ + P_ReturnThrustX(nil, finalAng, magnetPullForceScaled)
            magnetMomY = $ + P_ReturnThrustY(nil, finalAng, magnetPullForceScaled)

		    --electricity effect
		    for j = 1, MAGNET_ELECTRICITY_COUNT do
		        local elec = magnetInfo.electricity[j]
		        if not (elec and elec.valid) then
		            magnetInfo.electricity[j] = P_SpawnMobj(
						mobj.x,
						mobj.y,
						magnetElectricityVerticalOffsetOnPlayer(mobj),
						MT_DKRMAGNETELECTRICITY
					)
		            elec = magnetInfo.electricity[j]
		        end
		        updateMagnetismElectricityObject(elec)
		    end
            
			updateLoopingSound(magnetInfo.soundData, mobj)
            
            magnetInfo.timer = $ + 1
        end
    end
    
    if didDrag then
        mobj.momx = magnetMomX
        mobj.momy = magnetMomY
		-- repair properties related to FRICTION?????
		player.rmomx, player.rmomy = mobj.momx - player.cmomx, mobj.momy - player.cmomy
		-- repair speed since it was set earlier and won't be updated again
		player.speed = abs(FixedHypot(player.rmomx, player.rmomy))
    end
end

local function shouldCancelPrime(player)
    return player.playerstate ~= PST_LIVE
        or player.exiting
        or player.kartstuff[k_spinouttimer]
        or player.kartstuff[k_squishedtimer]
        or player.kartstuff[k_respawn]
end

local function doPlayerMobjThinkerStuff(mobj)
    if leveltime == 0 and not defrostingDebounce or mobj.player.spectator then
        resetPlayerVars(mobj.player)
    else
        if not mobj.player.dkrmagnetisminfo then
            resetPlayerVars(mobj.player)
        end
        
        if shouldCancelPrime(mobj.player) then
            local player = mobj.player
            player.dkrtarget = nil
            player.dkrtargetreticle = {}
			player.dkrmagnetprimed = false
        end

		if mobj.player.dkrmagnetprimed then
			refreshDKRTargetStuff(mobj.player)
		end

        updateMagnetDrags(mobj)
    end
end

addHook("MobjThinker", doPlayerMobjThinkerStuff, MT_PLAYER)

local function doDKRReticleThinkerStuff(mobj)
    local playerMo = mobj.tracer
    if not (playerMo and playerMo.valid) then
		P_RemoveMobj(mobj)
        return
    end
    
    local player = playerMo.player
    if not (player and player.valid)
        or not player.dkrtarget
		-- enter paranoia
        or not player.dkrtargetreticle.frame
        or not player.dkrtargetreticle.glass
        or player.spectator
        or shouldCancelPrime(player)
    then
        P_RemoveMobj(mobj)
    end
end

addHook("MobjThinker", doDKRReticleThinkerStuff, MT_DKRRETICLEFRAME)
addHook("MobjThinker", doDKRReticleThinkerStuff, MT_DKRRETICLEGLASS)

local function getElectricityScaleForProjectile(magnetProjectile)
	local result = FixedDiv(magnetProjectile.radius, mobjinfo[MT_DKRMAGNETELECTRICITY].radius)
	return result + (result >> 1)
end

local function magnetElectricityVerticalOffsetOnPlayer(playerMobj)
	return ternaryOperation(
            playerMobj.eflags & MFE_VERTICALFLIP,
            playerMobj.z + playerMobj.height - MAGNET_ELECTRICITY_OFFSET,
            playerMobj.z + MAGNET_ELECTRICITY_OFFSET
        )
end

local function magnetElectricityVerticalOffsetOnProjectile(magnetProjectile)
    return magnetProjectile.z +
        (magnetProjectile.height >> 2) +
        FixedMul(magnetProjectile.height >> 1, P_RandomFixed())
end

local function magnetElectricityAngleFromTarget(targetMobj, magnetProjectile)
	return R_PointToAngle2(targetMobj.x, targetMobj.y, magnetProjectile.x, magnetProjectile.y) +
		ANGLE_180 +
		FixedMul(ANG20, P_RandomFixed()) - ANG10
end

local function doDKRMagnetProjectileSpawnStuff(magnetProjectile)	
	-- scale to correct radius
	magnetProjectile.scale = FixedDiv(MAGNET_PROJECTILE_RADIUS, magnetProjectile.radius)

	-- the projectile needs electricity flying around it, like the pulling effect
	local electricity = {}
	
	for i = 1, MAGNET_ELECTRICITY_COUNT do
		local elec = P_SpawnMobj(
			magnetProjectile.x,
			magnetProjectile.y,
			magnetElectricityVerticalOffsetOnProjectile(magnetProjectile),
			MT_DKRMAGNETELECTRICITY
		)
		-- scale electricity to reflect size of projectile
		elec.scale = getElectricityScaleForProjectile(magnetProjectile)
		updateMagnetismElectricityObject(elec)
		electricity[i] = elec
	end
	-- have one electricity to indicate the target
	if magnetProjectile.tracer then
		local elec = P_SpawnMobj(
			magnetProjectile.tracer.x,
			magnetProjectile.tracer.y,
			magnetElectricityVerticalOffsetOnPlayer(magnetProjectile.tracer),
			MT_DKRMAGNETELECTRICITY
		)
		elec.angle = magnetElectricityAngleFromTarget(magnetProjectile.tracer, magnetProjectile)
		magnetProjectile.electricity.target = elec
	end
	magnetProjectile.electricity = electricity
	magnetProjectile.soundData = makeLoopingSoundData(magnetProjectile.info.attacksound, MAGNET_SOUND_LOOP)
	table.insert(dkrMagnetProjectiles, magnetProjectile)
end

addHook("MobjSpawn", doDKRMagnetProjectileSpawnStuff, MT_DKRMAGNETPROJECTILE)

local function handleMagnetProjectileMovement(magnetProjectile)
	local spookyGhost = P_SpawnGhostMobj(magnetProjectile)
	spookyGhost.flags2 = $ & ~MF2_DONTDRAW

	-- angle is unreliable, update it
	magnetProjectile.angle = R_PointToAngle2(0, 0, magnetProjectile.momx, magnetProjectile.momy)

	-- update angle to move slightly toward target IF there is one
	if magnetProjectile.tracer then
		local targetAngle = R_PointToAngle2(
			magnetProjectile.x,
			magnetProjectile.y,
			magnetProjectile.tracer.x,
			magnetProjectile.tracer.y
		)
		local desiredAngleTurn = targetAngle - magnetProjectile.angle
		-- only attempt to home into player if player is "ahead"
		if desiredAngleTurn >= ANGLE_270 and desiredAngleTurn <= ANGLE_90 then
			local maxTurnSpeed
			if magnetProjectile.tracer.player.kartstuff[k_curshield] == 1 then
				maxTurnSpeed = MAGNET_PROJECTILE_TURN_SPEED_MAX_ON_THUNDER_SHIELD
			else
				maxTurnSpeed = MAGNET_PROJECTILE_TURN_SPEED_MAX
			end
			if abs(desiredAngleTurn) > maxTurnSpeed then
				if desiredAngleTurn < 0 then
					desiredAngleTurn = -maxTurnSpeed
				else
					desiredAngleTurn = maxTurnSpeed
				end
			end
			magnetProjectile.angle = $ + desiredAngleTurn
		end
	end

	-- update momentum
	P_InstaThrust(magnetProjectile, magnetProjectile.angle, magnetProjectile.movefactor)
end

local function doDKRMagnetProjectileThinkerStuff(magnetProjectile)
	--[[
		based on source mobj variable use
		target = throwing player
		tracer = projectile's target
		threshold = timer to avoid self-collisions
		movefactor = current speed (because momx and momy are lossy)
	]]

	local aliveTime = -magnetProjectile.threshold + 10
	if aliveTime >= MAGNET_PROJECTILE_LIFETIME then
		P_KillMobj(magnetProjectile)
		return
	end

	-- makes sure owner is still valid
	if not magnetProjectile.targetPlayer.valid then
		magnetProjectile.targetPlayer = nil
		magnetProjectile.target = nil
	elseif not (magnetProjectile.target and magnetProjectile.target.valid) then
		-- refind the player's mobj
		if not magnetProjectile.targetPlayer.spectator and
			magnetProjectile.targetPlayer.mo and
			magnetProjectile.targetPlayer.mo.valid
		then
			magnetProjectile.target = magnetProjectile.targetPlayer.mo
		else
			-- we got a houdini
			magnetProjectile.target = nil
		end
		
	end
	-- make sure homing target is still valid
	if magnetProjectile.tracer and not magnetProjectile.tracer.valid then
		-- don't need to refind; just keep going
		magnetProjectile.tracer = nil
	end
	if aliveTime >= MAGNET_PROJECTILE_FORWARD_TIME then
		-- decelerate projectile
		magnetProjectile.movefactor = $ - MAGNET_PROJECTILE_DECELERATION
		if magnetProjectile.movefactor < 0 then
			magnetProjectile.movefactor = 0
		elseif magnetProjectile.movefactor == 0 then
			-- no need to process all the movement stuff
			magnetProjectile.momx, magnetProjectile.momy = 0, 0
		else
			handleMagnetProjectileMovement(magnetProjectile)
		end
	else
		handleMagnetProjectileMovement(magnetProjectile)
	end

	-- death sectors
	if P_CheckDeathPitCollide(magnetProjectile) then
		P_RemoveMobj(magnetProjectile)
		return
	end

	-- spring sectors, see p_mobj.c:7826
	if P_IsObjectOnGround(magnetProjectile) then
		local specialFofControlSector = P_ThingOnSpecial3DFloor(magnetProjectile)
		 -- on spring sector fof
		if specialFofControlSector and GetSecSpecial(specialFofControlSector.special, 3) == 1 or
			-- rigid test for if on main sector and it is a spring sector
			P_IsObjectOnRealGround(magnetProjectile, magnetProjectile.subsector.sector) and
				GetSecSpecial(magnetProjectile.subsector.sector.special, 3) == 1
		then
			K_DoPogoSpring(magnetProjectile, 0, 1)
		end
	end

	-- update electricity
	for i = 1, MAGNET_ELECTRICITY_COUNT do
		local elec = magnetProjectile.electricity[i]
		if not (elec and elec.valid) then
			magnetProjectile.electricity[i] = P_SpawnMobj(
				magnetProjectile.x,
				magnetProjectile.y,
				magnetElectricityVerticalOffsetOnProjectile(magnetProjectile),
				MT_DKRMAGNETELECTRICITY
			)
	        elec = mmagnetProjectile.electricity[i]
		end
		elec.scale = getElectricityScaleForProjectile(magnetProjectile)
		updateMagnetismElectricityObject(elec)
	end
	if magnetProjectile.tracer then
		local elec = magnetProjectile.electricity.target
		if not (elec and elec.valid) then
			elec = P_SpawnMobj(
				magnetProjectile.tracer.x,
				magnetProjectile.tracer.y,
				magnetElectricityVerticalOffsetOnPlayer(magnetProjectile.tracer),
				MT_DKRMAGNETELECTRICITY
			)
			magnetProjectile.electricity.target = elec
		end
		elec.scale = min(
			FixedDiv(
				magnetProjectile.movefactor,
				FixedMul(mobjinfo[MT_DKRMAGNETPROJECTILE].speed, magnetProjectileSpeedMult)
			),
			FRACUNIT
		)
	end

	updateLoopingSound(magnetProjectile.soundData, magnetProjectile)

	magnetProjectile.threshold = $ - 1
end

addHook("MobjThinker", doDKRMagnetProjectileThinkerStuff, MT_DKRMAGNETPROJECTILE)

local function mobjTypeIsDamageableDeflectorForMagnetProjectile(mobjType)
	return mobjType == MT_ORBINAUT or
		mobjType == MT_JAWZ or
		mobjType == MT_JAWZ_DUD or
		mobjType == MT_BANANA or
		mobjType == MT_BALLHOG or
		mobjType == MT_SSMINE
end

local function mobjTypeIsShieldDeflectorForMagnetProjectile(mobjType)
	return mobjType == MT_ORBINAUT_SHIELD or
		mobjType == MT_JAWZ_SHIELD or
		mobjType == MT_BANANA_SHIELD or
		mobjType == MT_SSMINE_SHIELD
end

local function spawnDroppedActiveItem(oldMobj, dropType, orbiting, orbitCenterMobj)
	local flipPolarity = P_MobjFlip(oldMobj.target)

	local dropMobj = P_SpawnMobj(oldMobj.x, oldMobj.y, oldMobj.z, dropType)
	dropMobj.target = oldMobj.target
	dropMobj.angle = oldMobj.angle
	dropMobj.flags2 = oldMobj.flags2
	dropMobj.flags = $ | MF_NOCLIPTHING
	--dropMobj.floorz = oldMobj.floorz
	--dropMobj.ceilingz = oldMobj.ceilingz
	-- if these can't be set, then just use the values of oldMobj later

	if P_IsObjectOnGround(dropMobj.target) then
		if flipPolarity == 1 then
			if oldMobj.floorz > dropMobj.target.z - dropMobj.height then
				dropMobj.z = oldMobj.floorz
			end
		else
			if oldMobj.ceilingz < dropMobj.target.z + dropMobj.target.height + dropMobj.height then
				dropMobj.z = oldMobj.ceilingz - dropMobj.height
			end
		end
	end

	if orbiting then
		dropMobj.flags2 = $ | MF2_AMBUSH
		dropMobj.z = $ + flipPolarity
		dropMobj.momx = orbitCenterMobj.momx >> 1
		dropMobj.momy = orbitCenterMobj.momy >> 1
		dropMobj.momz = 3 * flipPolarity * mapobjectscale
		if dropMobj.eflags & MFE_UNDERWATER then
			dropMobj.momz = $ * 117 / 200
		end
		P_Thrust(dropMobj, oldMobj.angle - ANGLE_90, 6 * mapobjectscale)
		dropMobj.movecount = 2
		dropMobj.movedir = oldMobj.angle - ANGLE_90
		dropMobj.state = dropMobj.info.deathstate
		dropMobj.tics = -1
		if dropType == MT_JAWZ_DUD then
			dropMobj.z = $ + 20 * flipPolarity * dropMobj.scale
		else
			dropMobj.color = oldMobj.color
			dropMobj.angle = $ - ANGLE_90
		end
	else
		dropMobj.flags = $ & ~MF_NOCLIPTHING
		dropMobj.threshold = 10
	end
end

local function removeOrbitingItem(player, orbitingItem)
	player.kartstuff[k_itemamount] = $ - 1
	if player.kartstuff[k_itemamount] <= 0 then
		-- just in case
		player.kartstuff[k_itemamount] = 0
		player.kartstuff[k_itemtype] = KITEM_NONE
		player.kartstuff[k_itemheld] = 0
		player.mo.hnext = nil
	end
	K_RepairOrbitChain(orbitingItem)
	P_RemoveMobj(orbitingItem)
end

local function removeHeldItemsFromSlot(player, removeCount)
	player.kartstuff[k_itemamount] = $ - removeCount
	if player.kartstuff[k_itemamount] <= 0 then
		-- just in case
		player.kartstuff[k_itemamount] = 0
		player.kartstuff[k_bananadrag] = 0
		player.kartstuff[k_itemheld] = 0
		player.kartstuff[k_itemtype] = KITEM_NONE
	end
end

local function doDKRMagnetProjectileCollideStuff(magnetProjectile, overlappingMobj)
	-- I don't know how it's possible but it is
	if not (magnetProjectile and magnetProjectile.valid and overlappingMobj and overlappingMobj.valid) then
		return false
	end

	-- general vertical check
	if magnetProjectile.z > overlappingMobj.z + overlappingMobj.height or
		magnetProjectile.z + magnetProjectile.height < overlappingMobj.z
	then
		return false
	end

	-- we don't care about stuff that's probably dead
	if magnetProjectile.health <= 0 or overlappingMobj.health <= 0 then
		return false
	end

	local collidingType = overlappingMobj.type
	if collidingType == MT_PLAYER then
		local collidingPlayer = overlappingMobj.player
		-- don't collide with invalid players
		if not (collidingPlayer and collidingPlayer.valid) then
			return false
		end
		local collidingKartStuff = collidingPlayer.kartstuff
		-- don't collide with...
			-- spectating players
		if collidingPlayer.spectator or
			-- not alive players
			collidingPlayer.playerstate ~= PST_LIVE or
			-- respawning players
			collidingKartStuff[k_respawn] > 0 or
			-- players waiting to come back in battle (NOT COMEBACK KARTS)
			G_BattleGametype() and collidingKartStuff[k_comebacktimer] > 0
		then
			return false
		end

		-- is this the user?
		if magnetProjectile.target and collidingPlayer == magnetProjectile.target.player then
			-- try to pick it back up!
			local collidingXItemData = collidingPlayer.xItemData
			-- can the player even pick it up? Don't pick up if...
				-- xItem data isn't initialized
			if not collidingXItemData or
				-- no bumpers in battle
				G_BattleGametype() and collidingKartStuff[k_bumper] <= 0 or
				-- hasn't been enough time since use
				magnetProjectile.threshold > 0 or
				-- item routlette is spinning
				collidingXItemData.xItem_roulette > 0 or
				-- item slot is locked (?)
				collidingXItemData.xItem_itemSlotLocked or
				-- hyudoro is stealing an item for you
				collidingKartStuff[k_stealingtimer] > 0 or
				-- hyudoro is stealing an item from you
				collidingKartStuff[k_stolentimer] > 0 or
				-- you are grown (talk with KRU about this)
				collidingKartStuff[k_growshrinktimer] > 0 or
				-- you have rocket sneakers
				collidingKartStuff[k_rocketsneakertimer] > 0 or
				-- you ran into an eggman monitor
				collidingKartStuff[k_eggmanexplode] > 0 or
				-- you have a different item occupying the slot
				collidingKartStuff[k_itemtype] and
					collidingKartStuff[k_itemtype] ~= DKR_BACK_MAGNET_ID
			then
				return false
			end
			
			collidingKartStuff[k_itemtype] = DKR_BACK_MAGNET_ID
			collidingKartStuff[k_itemamount] = $ + 1
			-- make the noise
			S_StartSound(overlappingMobj, sfx_itpick)

			-- don't kill or it'll make the clash!
			P_RemoveMobj(magnetProjectile)
		else
			-- don't collide with players using hyudoro (you pretty much collide with nothing anyway)
			if collidingKartStuff[k_hyudorotimer] > 0 then
				return false
			end
			-- NOTE: I-frames will not save you from the magnet's wrath

			local ownerMobj = magnetProjectile.target
			-- can't find the user? Well that's a problem now isn't it
			if not ownerMobj or playerKillsMagnetProjectile(ownerMobj.player) then
				-- die
				P_KillMobj(magnetProjectile)
			else
				local ownerPlayer = ownerMobj.player
				--you're going to brazil, and I'm brazil. C'mere
				table.insert(collidingPlayer.dkrmagnetisminfo, {
					timer = 0,
					timeUp = MAGNET_TIME,
					target = ownerPlayer,
					angRespect = ownerPlayer,
					electricity = {},
					soundData = makeLoopingSoundData(sfx_atdrmg, MAGNET_SOUND_LOOP)
				})

				-- have a thunder shield? Oh, buddy...
				if collidingKartStuff[k_curshield] == 1 then
					-- no more shield for you noob
					collidingKartStuff[k_itemamount] = 0
					collidingKartStuff[k_itemtype] = KITEM_NONE
				end
				
				--hitfeed compat inserted by mine
				if hitfeed then
					local source = generateHFForPlayer(ownerPlayer)
					local victim = generateHFForPlayer(collidingPlayer)
					HF_SendHitMessage(source, victim, "H_DKRBM")
				end

				--fes compat also by mine
				if fesScoring then
					FES_scoring_rewardPlayerAction(ownerPlayer, "xitem_dkrmagnet")
				end
				
				-- don't kill or it'll make the clash!
				P_RemoveMobj(magnetProjectile)
			end
		end
	elseif mobjTypeIsDamageableDeflectorForMagnetProjectile(collidingType) then
		-- mimic source's code for the objects EXACTLY
		-- see p_map.c:789
		if overlappingMobj.type ~= MT_SSMINE then
			if overlappingMobj.eflags & MFE_VERTICALFLIP then
				overlappingMobj.z = $ - overlappingMobj.height
			else
				overlappingMobj.z = $ + overlappingMobj.height
			end
			S_StartSound(overlappingMobj, overlappingMobj.info.deathsound)
			P_SetObjectMomZ(overlappingMobj, 8 << FRACBITS, false)
			P_InstaThrust(
				overlappingMobj,
				R_PointToAngle2(magnetProjectile.x, magnetProjectile.y, overlappingMobj.x, overlappingMobj.y) + ANGLE_90,
				16 << FRACUNIT
			)
		end
		P_KillMobj(magnetProjectile, overlappingMobj, overlappingMobj)
		P_KillMobj(overlappingMobj, magnetProjectile, magnetProjectile)
	elseif mobjTypeIsShieldDeflectorForMagnetProjectile(collidingType) then
		-- shielding items drop as if the racer took damage
		-- again, mimic source. See k_kart.c:3805
		-- only work on this item if its owner still exists
		local targetPlayerMobj = overlappingMobj.target
		if targetPlayerMobj and targetPlayerMobj.valid then
			-- do not collide if owners match
			if targetPlayerMobj == magnetProjectile.target then
				return false
			end
			P_KillMobj(magnetProjectile, overlappingMobj, overlappingMobj)
			local targetPlayer = targetPlayerMobj.player
			if targetPlayer and targetPlayer.valid then
				if collidingType == MT_ORBINAUT_SHIELD then
					spawnDroppedActiveItem(overlappingMobj, MT_ORBINAUT, true, targetPlayerMobj)
					removeOrbitingItem(targetPlayer, overlappingMobj)
				elseif collidingType == MT_JAWZ_SHIELD then
					spawnDroppedActiveItem(overlappingMobj, MT_JAWZ_DUD, true, targetPlayerMobj)
					removeOrbitingItem(targetPlayer, overlappingMobj)
				elseif collidingType == MT_BANANA_SHIELD then
					-- normally the entire chain gets deleted unless the tip is hit
					-- here, only from the hit banana backward is cut off
					-- first update the previous banana since the mobj referncing it will be removed
					if overlappingMobj.hprev and overlappingMobj.hprev.valid then
						overlappingMobj.hprev.hnext = nil
						if targetPlayerMobj.hnext == overlappingMobj.hprev then
							targetPlayerMobj.hnext = nil
						end
					end
					-- remove each banana backwards
					-- store reference since mobjs will be removed
					local targetHNext = overlappingMobj
					local cachedHNext = targetHNext
					-- count how many are lost
					local droppedCount = 0
					while cachedHNext and cachedHNext.valid do
						targetHNext = cachedHNext
						cachedHNext = targetHNext.hnext
						spawnDroppedActiveItem(targetHNext, MT_BANANA, false)
						-- remove old object
						P_RemoveMobj(targetHNext)
						droppedCount = $ + 1
					end
					-- update item slot info
					removeHeldItemsFromSlot(targetPlayer, droppedCount)
				elseif collidingType == MT_SSMINE_SHIELD then
					spawnDroppedActiveItem(overlappingMobj, MT_SSMINE, false)
					targetPlayerMobj.hnext = nil
					-- update item slot info
					removeHeldItemsFromSlot(targetPlayer, 1)
					-- remove old object
					P_RemoveMobj(overlappingMobj)
				end
			end
		end
	elseif overlappingMobj.flags & MF_SPRING then
		-- boi-oi-oi-oing
		P_DoSpring(overlappingMobj, magnetProjectile)
	end

	-- never actually collide with something or it'll be blocked
	return false
end

addHook("MobjCollide", doDKRMagnetProjectileCollideStuff, MT_DKRMAGNETPROJECTILE)
addHook("MobjMoveCollide", doDKRMagnetProjectileCollideStuff, MT_DKRMAGNETPROJECTILE)

local function doDKRMagnetProjectileDeathStuff(magnetProjectile)
	-- TODO: this really needs its own thing; be sure to add clash where needed if changed
	local clashMobj = P_SpawnMobj(
		magnetProjectile.x,
		magnetProjectile.y,
		magnetProjectile.z + (magnetProjectile.height >> 1),
		MT_ITEMCLASH
	)
end

addHook("MobjDeath", doDKRMagnetProjectileDeathStuff, MT_DKRMAGNETPROJECTILE)

local function doDKRMagnetProjectileRemovedStuff(magnetProjectile)
	for _, elec in pairs(magnetProjectile.electricity) do
		P_KillMobj(elec)
	end
	for i, storedProjectile in ipairs(dkrMagnetProjectiles) do
		if storedProjectile == magnetProjectile then
			table.remove(dkrMagnetProjectiles, i)
			break
		end
	end
end

addHook("MobjRemoved", doDKRMagnetProjectileRemovedStuff, MT_DKRMAGNETPROJECTILE)

local function updatePerRaceScaledVars()
    targetingDistThresScaled = getMapObjectScaled(TARGETING_DIST_THRES)
    magnetKillDistThresScaled = getMapObjectScaled(MAGNET_KILL_DIST_THRES)
    magnetTargetOffsetDistScaled = getMapObjectScaled(MAGNET_TARGET_OFFSET_DIST)
    magnetPullForceScaled = getMapObjectScaled(MAGNET_PULL_FORCE)
	magnetProjectileSpeedMult = getMapObjectScaled(MAGNET_PROJECTILE_SPEED_MULT_FOR_KART_SPEED_LOOKUP[gamespeed])
end

local function doThinkFrameStuff()
    if leveltime == 0
        if not defrostingDebounce then
            updatePerRaceScaledVars()
			dkrMagnetProjectiles = {}
            defrostingDebounce = true
        end
    else
        if leveltime == 1 then
            defrostingDebounce = false
        end
        for player in players.iterate do
            --very confident if player.dkrmagnetisminfo is nonnil, player is valid for this 
            if player.dkrmagnetisminfo then
                local playerMo = player.mo
				local elecX = playerMo.x
				local elecY = playerMo.y
				local elecZ = magnetElectricityVerticalOffsetOnPlayer(playerMo)
                for _, magnetInfo in ipairs(player.dkrmagnetisminfo) do
                    for i = 1, MAGNET_ELECTRICITY_COUNT do
						local elec = magnetInfo.electricity[i]
						if not (elec and elec.valid) then
							elec = P_SpawnMobj(elecX, elecY, elecZ, MT_DKRMAGNETELECTRICITY)
							magnetInfo.electricity[i] = elec
						else
		                    P_MoveOrigin(elec, elecX, elecY, elecZ)
						end
                    end
                end
            end
        end
		for i, magnetProjectile in ipairs(dkrMagnetProjectiles) do
			local elecX = magnetProjectile.x
			local elecY = magnetProjectile.y
			for i = 1, MAGNET_ELECTRICITY_COUNT do
				local elec = magnetProjectile.electricity[i]
				local elecZ = magnetElectricityVerticalOffsetOnProjectile(magnetProjectile)
				if not (elec and elec.valid) then
					elec = P_SpawnMobj(elecX, elecY, elecZ, MT_DKRMAGNETELECTRICITY)
					magnetProjectile.electricity[i] = elec
					elec.scale = getElectricityScaleForProjectile(magnetProjectile)
				else
					P_MoveOrigin(elec, elecX, elecY, elecZ)
				end
			end
			if magnetProjectile.tracer then
				if not magnetProjectile.tracer.valid then
					magnetProjectile.tracer = nil
				else
					local elec = magnetProjectile.electricity.target
					local elecX = magnetProjectile.tracer.x
					local elecY = magnetProjectile.tracer.y
					local elecZ = magnetElectricityVerticalOffsetOnPlayer(magnetProjectile.tracer)
					if not (elec and elec.valid) then
						elec = P_SpawnMobj(elecX, elecY, elecZ, MT_DKRMAGNETELECTRICITY)
						magnetProjectile.electricity.target = elec
					else
						P_MoveOrigin(elec, elecX, elecY, elecZ)
					end
					elec.angle = magnetElectricityAngleFromTarget(magnetProjectile.tracer, magnetProjectile)
				end
			end
		end
    end
end

addHook("ThinkFrame", doThinkFrameStuff)

-- this is incredibly jank
local function checkThunderShieldMagnetProjectileDamage(usingPlayer, usingPlayerCmd)
	-- mimic P_NukeEnemies; see p_user.c:6731
	local playerMobj = usingPlayer.mo
	local scaledRadius = FixedMul(THUNDER_SHIELD_DAMAGE_RADIUS, mapobjectscale)
	-- magnetProjectiles can be removed from the cache table on death, so iterate accordingly
	for i = #dkrMagnetProjectiles, 1, -1 do
		local magnetProjectile = dkrMagnetProjectiles[i]
		if abs(playerMobj.x - magnetProjectile.x) <= THUNDER_SHIELD_DAMAGE_RADIUS and
			abs(playerMobj.y - magnetProjectile.y) <= THUNDER_SHIELD_DAMAGE_RADIUS and
			abs(playerMobj.z - magnetProjectile.z) <= THUNDER_SHIELD_DAMAGE_RADIUS and
			FixedHypot(
				FixedHypot(playerMobj.x - magnetProjectile.x, playerMobj.y - magnetProjectile.y),
				playerMobj.z - magnetProjectile.z
			) <= radius
		then
			-- not today, magnets
			-- DO NOT REMOVE FROM TABLE; kill function does it
			P_KillMobj(magnetProjectile)
		end
	end
end
local xItemThunderShieldModData = {usefunc = checkThunderShieldMagnetProjectileDamage}
-- NOTE: namespace here is not the same namespace as below
xItemLib.func.setXItemModData(DKR_BACK_MAGNET_MOD_NAMESPACE, KITEM_THUNDERSHIELD, xItemThunderShieldModData)

local function drawDKRBackMagnet(v, p, c)
    xItemLib.func.hudDrawItem(v, p, c, nil, p.dkrmagnetprimed and 2 or nil)
end

if fesScoring then
	FES_scoring_addAction("xitem_dkrmagnet", "Magnet Accuracy", 25)
end

updatePerRaceScaledVars()
for player in players.iterate do
    resetPlayerVars(player)
end

xItemLib.func.addItem(
    DKR_BACK_MAGNET_ITEM_NAMESPACE,
    "DKR Self-Magnet",
    "ATIDKRBM",
    "ATIDRBMS",
    0,
    {0, 0, 1, 3, 2, 2, 0, 0, 0, 0},
    {2, 0},
    nil,
    useDKRBackMagnet,
    drawDKRBackMagnet,
    nil,
    nil,
    {0, {SPR_XITS, 4}},
    true,
    preuseDKRBackMagnet
)

DKR_BACK_MAGNET_ID = xItemLib.func.getItemDataByName(DKR_BACK_MAGNET_ITEM_NAMESPACE).id

xItemLib.func.setXItemModData("XITEM_DESCRIPTIONS", DKR_BACK_MAGNET_ITEM_NAMESPACE, {
	tipDescription1 = "Fires homing magnetism at target",
	tipDescription2 = "Successful hit harshly drags victim backwards",
	tipUsage1 = "Tap or hold \130Fire\128 to lock on",
	tipUsage2 = "If tapped, tap again, or if held, release to use",
})
